#include <ToppersASP.h>
#include <ToppersASP_FreeRTOS_API.h>
#include <AzureIoTHub.h>
#include <Wire.h>
#include <ZumoShieldN.h>
#include <stdio.h>
#include <stdlib.h>

#include <rpcWiFi.h>

using namespace erpc;

#include "iot_configs.h"
#include "sample_init.h"

#ifdef SAMPLE_MQTT
#include "AzureIoTProtocol_MQTT.h"
#include "iothubtransportmqtt.h"
#endif // SAMPLE_MQTT
#ifdef SAMPLE_HTTP
#include "AzureIoTProtocol_HTTP.h"
#include "iothubtransporthttp.h"
#endif // SAMPLE_HTTP

#include "Free_Fonts.h" // Include the header file attached to this sketch

#include <TFT_eSPI.h> // Hardware-specific library
#include <SPI.h>

TFT_eSPI tft = TFT_eSPI();                   // Invoke custom library with default width and height

unsigned long drawTime = 0;

static const char ssid[] = IOT_CONFIG_WIFI_SSID;
static const char pass[] = IOT_CONFIG_WIFI_PASSWORD;

/* Define several constants/global variables */
static const char* connectionString = DEVICE_CONNECTION_STRING;
static bool g_continueRunning = true;
static size_t g_message_count_send_confirmations = 0;
static bool g_run_demo = false;

IOTHUB_MESSAGE_HANDLE message_handle;
size_t messages_sent = 0;
#define MESSAGE_COUNT 1
char telemetry_msg[] = "-01234";
const char* quit_msg = "quit";
const char* exit_msg = "exit";

IOTHUB_DEVICE_CLIENT_LL_HANDLE device_ll_handle;

static int callbackCounter;
int receiveContext = 0;

char log_buf[1024];

const int xpos =  0;
const int ypos = 40;

char motor_control[20];
int n = 0;

#define REFLECTANCE_THRESHOLD  400

void log_function(LOG_CATEGORY log_category, const char* file, const char* func, int line, unsigned int options, const char* format, ...)
{
  va_list args;
  va_start( args, format );
  sprintf(log_buf, "info %s: func %s line %d ", file, func, line);
  Serial.print(log_buf);
  vsprintf(log_buf, format, args);
  Serial.println(log_buf);
  va_end( args );
}

/* -- receive_message_callback --
   Callback method which executes upon receipt of a message originating from the IoT Hub in the cloud.
   Note: Modifying the contents of this method allows one to command the device from the cloud.
*/
static IOTHUBMESSAGE_DISPOSITION_RESULT receive_message_callback(IOTHUB_MESSAGE_HANDLE message, void* userContextCallback)
{
  Serial.println("receive_message_callback\n");
  int* counter = (int*)userContextCallback;
  const unsigned char* buffer;
  size_t size;
  const char* messageId;

  // Message properties
  if ((messageId = IoTHubMessage_GetMessageId(message)) == NULL)
  {
    messageId = "<null>";
  }

  // Message content
  if (IoTHubMessage_GetByteArray(message, (const unsigned char**)&buffer, &size) != IOTHUB_MESSAGE_OK)
  {
    LogInfo("unable to retrieve the message data\r\n");
  }
  else
  {
    LogInfo("Received Message [%d]\r\n Message ID: %s\r\n Data: <<<%.*s>>> & Size=%d\r\n", *counter, messageId, (int)size, buffer, (int)size);
    // If we receive the word 'quit' then we stop running
    if (size == (strlen(quit_msg) * sizeof(char)) && memcmp(buffer, quit_msg, size) == 0)
    {
      g_continueRunning = false;
    }
  }

  /* Some device specific action code goes here... */
  (*counter)++;
  return IOTHUBMESSAGE_ACCEPTED;
}

/* -- connection_status_callback --
   Callback method which executes on receipt of a connection status message from the IoT Hub in the cloud.
*/
static void connection_status_callback(IOTHUB_CLIENT_CONNECTION_STATUS result, IOTHUB_CLIENT_CONNECTION_STATUS_REASON reason, void* user_context)
{
  //Serial.println("connection_states_callback\n");
  (void)reason;
  (void)user_context;
  // This sample DOES NOT take into consideration network outages.
  if (result == IOTHUB_CLIENT_CONNECTION_AUTHENTICATED)
  {
    LogInfo("The device client is connected to iothub\r\n");
  }
  else
  {
    LogInfo("The device client has been disconnected\r\n");
  }
}

/* -- send_confirm_callback --
   Callback method which executes upon confirmation that a message originating from this device has been received by the IoT Hub in the cloud.
*/
static void send_confirm_callback(IOTHUB_CLIENT_CONFIRMATION_RESULT result, void* userContextCallback)
{
  //Serial.println("send_confirm_callback\n");
  (void)userContextCallback;
  // When a message is sent this callback will get envoked
  g_message_count_send_confirmations++;
  LogInfo("Confirmation callback received for message %lu with result %s\r\n", (unsigned long)g_message_count_send_confirmations, MU_ENUM_TO_STRING(IOTHUB_CLIENT_CONFIRMATION_RESULT, result));
}

/* -- reset_esp_helper --
   waits for call of exit_msg over Serial line to reset device
*/
static void reset_esp_helper()
{
#ifdef is_esp_board
  // Read from local serial
  if (Serial.available()) {
    String s1 = Serial.readStringUntil('\n');// s1 is String type variable.
    Serial.print("Received Data: ");
    Serial.println(s1);//display same received Data back in serial monitor.

    // Restart device upon receipt of 'exit' call.
    int e_start = s1.indexOf('e');
    String ebit = (String) s1.substring(e_start, e_start + 4);
    if (ebit == exit_msg)
    {
      ESP.restart();
    }
  }
#endif // is_esp_board
}

void front_trace() {
  while ((reflectances.value(1) < REFLECTANCE_THRESHOLD) ||
         (reflectances.value(6) < REFLECTANCE_THRESHOLD)) {
    if (reflectances.value(2) > REFLECTANCE_THRESHOLD) {
      motors.setSpeeds(-100, 100);
    }
    else if (reflectances.value(5) > REFLECTANCE_THRESHOLD) {
      motors.setSpeeds(100, -100);
    }
    else {
      motors.setSpeeds(100, 100);
    }
    reflectances.update();
    imu.turnSensorUpdate();
  }
  if (reflectances.value(1) > REFLECTANCE_THRESHOLD) {
    while (reflectances.value(6) < REFLECTANCE_THRESHOLD) {
      motors.setSpeeds(0, 100);
      imu.turnSensorUpdate();
    }
  }
  if (reflectances.value(6) > REFLECTANCE_THRESHOLD) {
    while (reflectances.value(1) < REFLECTANCE_THRESHOLD) {
      motors.setSpeeds(100, 0);
      imu.turnSensorUpdate();
    }
  }
  motors.setSpeeds(100, 100);
  delay(360);
  motors.setSpeeds(0, 0);
}

extern void _user_inirtn(void);

void setup() {
  Serial1.begin(115200);
  delay(1000);

  //xlogging_set_log_function(log_function);

  pinMode(WIO_5S_UP, INPUT_PULLUP);
  pinMode(WIO_5S_DOWN, INPUT_PULLUP);
  pinMode(WIO_5S_LEFT, INPUT_PULLUP);
  pinMode(WIO_5S_RIGHT, INPUT_PULLUP);
  pinMode(WIO_5S_PRESS, INPUT_PULLUP);

  tft.begin();
  tft.fillScreen(TFT_BLACK); // Clear screen to navy background

  imu.begin();
  imu.configureForTurnSensing();
  imu.turnSensorReset();

  // For comaptibility with Adafruit_GFX library the text background is not plotted when using the print class
  // even if we specify it.
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  tft.setCursor(xpos, ypos);    // Set cursor near top left corner of screen

  tft.setFreeFont(FSB18);       // Select Free Serif 24 point font
  tft.println();                // Move cursor down a line
  tft.println("Connecting"); // Print the font name onto the TFT screen
  tft.println("WiFi..");

  // Select the Protocol to use with the connection
#ifdef SAMPLE_MQTT
  IOTHUB_CLIENT_TRANSPORT_PROVIDER protocol = MQTT_Protocol;
#endif // SAMPLE_MQTT
#ifdef SAMPLE_HTTP
  IOTHUB_CLIENT_TRANSPORT_PROVIDER protocol = HTTP_Protocol;
#endif // SAMPLE_HTTP

  sample_init(ssid, pass);
  _user_inirtn();

  // Used to initialize IoTHub SDK subsystem
  (void)IoTHub_Init();
  // Create the iothub handle here
  device_ll_handle = IoTHubDeviceClient_LL_CreateFromConnectionString(connectionString, protocol);
  LogInfo("Creating IoTHub Device handle\r\n");

  if (device_ll_handle == NULL)
  {
    LogInfo("Error AZ002: Failure creating Iothub device. Hint: Check you connection string.\r\n");
  }
  else
  {
    // Set any option that are neccessary.
    // For available options please see the iothub_sdk_options.md documentation in the main C SDK
    // turn off diagnostic sampling
    int diag_off = 0;
    IoTHubDeviceClient_LL_SetOption(device_ll_handle, OPTION_DIAGNOSTIC_SAMPLING_PERCENTAGE, &diag_off);

#ifndef SAMPLE_HTTP
    // Example sdk status tracing for troubleshooting
    bool traceOn = true;
    IoTHubDeviceClient_LL_SetOption(device_ll_handle, OPTION_LOG_TRACE, &traceOn);
#endif // SAMPLE_HTTP

    // Setting the Trusted Certificate.
    IoTHubDeviceClient_LL_SetOption(device_ll_handle, OPTION_TRUSTED_CERT, certificates);

#if defined SAMPLE_MQTT
    //Setting the auto URL Encoder (recommended for MQTT). Please use this option unless
    //you are URL Encoding inputs yourself.
    //ONLY valid for use with MQTT
    bool urlEncodeOn = true;
    IoTHubDeviceClient_LL_SetOption(device_ll_handle, OPTION_AUTO_URL_ENCODE_DECODE, &urlEncodeOn);
    /* Setting Message call back, so we can receive Commands. */
    if (IoTHubClient_LL_SetMessageCallback(device_ll_handle, receive_message_callback, &receiveContext) != IOTHUB_CLIENT_OK)
    {
      LogInfo("ERROR: IoTHubClient_LL_SetMessageCallback..........FAILED!\r\n");
    }
#endif // SAMPLE_MQTT

    // Setting connection status callback to get indication of connection to iothub
    (void)IoTHubDeviceClient_LL_SetConnectionStatusCallback(device_ll_handle, connection_status_callback, NULL);
  }
}

int loop_count;
extern void run_demo(intptr_t exinf);

void loop() {

  if (g_run_demo)
  {
    run_demo();
    g_run_demo = false;
  }
  reset_esp_helper();
}

/*
    サービスコールのエラーのログ出力
*/
void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
  if (ercd < 0) {
    t_perror(LOG_ERROR, file, line, expr, ercd);
  }
}

#define SVC_PERROR(expr)  svc_perror(__FILE__, __LINE__, #expr, (expr))

/*
    各タスクの優先度の定義
*/

#define MAIN_PRIORITY  5   /* メインタスクの優先度 */
#define HIGH_PRIORITY 9   /* 並行実行されるタスクの優先度 */
#define MID_PRIORITY  10
#define LOW_PRIORITY  11

/*
    ターゲットに依存する可能性のある定数の定義
*/
#ifndef TASK_PORTID
#define TASK_PORTID   1     /* 文字入力するシリアルポートID */
#endif /* TASK_PORTID */

#ifndef STACK_SIZE
#define STACK_SIZE    1024    /* タスクのスタックサイズ */
#endif /* STACK_SIZE */

#ifndef LOOP_REF
#define LOOP_REF    ULONG_C(1000000)  /* 速度計測用のループ回数 */
#endif /* LOOP_REF */

extern void task(intptr_t exinf);
extern void runZumo(intptr_t exinf);
extern void run_demo(intptr_t exinf);
extern void main_task(intptr_t exinf);

#define MAIN_TASK TASK10

void
_user_inirtn(void)
{
  T_CTSK	ctsk;
  ER		ercd;

  //  ctsk.tskatr = TA_NULL;
  //  ctsk.exinf = 1;
  //  ctsk.task = run_demo;
  //  ctsk.itskpri = HIGH_PRIORITY;
  //  ctsk.stksz = STACK_SIZE;
  //  ctsk.stk = NULL;
  //  ercd = cre_tsk(TASK9, &ctsk);
  //  assert(ercd == E_OK);

  ctsk.tskatr = TA_NULL;
  ctsk.exinf = 2;
  ctsk.task = runZumo;
  ctsk.itskpri = MID_PRIORITY;
  ctsk.stksz = STACK_SIZE;
  ctsk.stk = NULL;
  ercd = cre_tsk(TASK8, &ctsk);
  assert(ercd == E_OK);

  ctsk.tskatr = TA_ACT;
  ctsk.exinf = 0;
  ctsk.task = main_task;
  ctsk.itskpri = LOW_PRIORITY;
  ctsk.stksz = STACK_SIZE;
  ctsk.stk = NULL;
  ercd = cre_tsk(MAIN_TASK, &ctsk);
  assert(ercd == E_OK);
}

/*
    並行実行されるタスクへのメッセージ領域
*/
char	message[3];

/*
    ループ回数
*/
ulong_t	task_loop;		/* タスク内でのループ回数 */
ulong_t	tex_loop;		/* 例外処理ルーチン内でのループ回数 */

/*
    並行実行されるタスク
*/
void task1(intptr_t exinf)
{
  volatile ulong_t	i;
  int_t		n = 0;
  int_t		tskno = (int_t) exinf;
  const char	*graph[] = { "|", "  +", "    *" };
  char		c;

  SVC_PERROR(ena_tex());
  while (true) {
    //syslog(LOG_NOTICE, "task%d is running (%03d).   %s", tskno, ++n, graph[tskno - 1]);
    Serial.print("task1 is running\n");
    for (i = 0; i < task_loop; i++);
    c = message[tskno - 1];
    message[tskno - 1] = 0;
  }
}

void task2(intptr_t exinf)
{
  volatile ulong_t  i;
  int_t   n = 0;
  int_t   tskno = (int_t) exinf;
  const char  *graph[] = { "|", "  +", "    *" };
  char    c;

  SVC_PERROR(ena_tex());
  while (true) {
    //syslog(LOG_NOTICE, "task%d is running (%03d).   %s", tskno, ++n, graph[tskno - 1]);
    Serial.print("task2 is running");
    for (i = 0; i < task_loop; i++);
    c = message[tskno - 1];
    message[tskno - 1] = 0;
  }
}

void runZumo(intptr_t exinf) {
  syslog(LOG_NOTICE, "runZumo");
  reflectances.update();

  if (motor_control[n] == 'f') {
    tft.print("f");
    front_trace();
  }
  else if (motor_control[n] == 'b') {
    tft.print("b");
    while ((reflectances.value(1) < REFLECTANCE_THRESHOLD &&
            reflectances.value(2) < REFLECTANCE_THRESHOLD) ||
           (reflectances.value(6) < REFLECTANCE_THRESHOLD &&
            reflectances.value(5) < REFLECTANCE_THRESHOLD)) {
      reflectances.update();
      imu.turnSensorUpdate();
      if (reflectances.value(2) > REFLECTANCE_THRESHOLD) {
        motors.setSpeeds(100, -100);
      }
      else if (reflectances.value(5) > REFLECTANCE_THRESHOLD) {
        motors.setSpeeds(-100, 100);
      }
      else {
        motors.setSpeeds(-100, -100);
      }
    }
    motors.setSpeeds(100, 100);
    delay(300);
    motors.setSpeeds(0, 0);
  }
  else if (motor_control[n] == 'r') {
    tft.print("r");
    reflectances.update();
    while (reflectances.value(6) < REFLECTANCE_THRESHOLD) {
      motors.setSpeeds(100, -100);
      reflectances.update();
      imu.turnSensorUpdate();
    }
    while (reflectances.value(3) < REFLECTANCE_THRESHOLD) {
      motors.setSpeeds(100, -100);
      reflectances.update();
      imu.turnSensorUpdate();
    }

    front_trace();

  }
  else if (motor_control[n] == 'l') {
    tft.print("l");
    reflectances.update();
    while (reflectances.value(1) < REFLECTANCE_THRESHOLD) {
      motors.setSpeeds(-100, 100);
      reflectances.update();
      imu.turnSensorUpdate();
    }
    while (reflectances.value(4) < REFLECTANCE_THRESHOLD) {
      motors.setSpeeds(-100, 100);
      reflectances.update();
      imu.turnSensorUpdate();
    }

    front_trace();
  }

  n++;

  syslog(LOG_NOTICE, "%s", motor_control);

  sprintf(telemetry_msg, "%d", imu.turnSensorAngleDegree());
  syslog(LOG_NOTICE, "%s", telemetry_msg);

  g_run_demo = true;
  SVC_PERROR(ext_tsk());
}

/* -- run_demo --
   Runs active task of sending telemetry to IoTHub
   WARNING: only call this function once, as it includes steps to destroy handles and clean up at the end.
*/
static void run_demo()
{
  int result = 0;

  syslog(LOG_NOTICE, "run_demo\n");

  // action phase of the program, sending messages to the IoT Hub in the cloud.
  do
  {
    if (messages_sent < MESSAGE_COUNT)
    {

      // Construct the iothub message from a string or a byte array
      message_handle = IoTHubMessage_CreateFromString(telemetry_msg);
      //message_handle = IoTHubMessage_CreateFromByteArray((const unsigned char*)msgText, strlen(msgText)));

      // Set Message property
      /*(void)IoTHubMessage_SetMessageId(message_handle, "MSG_ID");
        (void)IoTHubMessage_SetCorrelationId(message_handle, "CORE_ID");
        (void)IoTHubMessage_SetContentTypeSystemProperty(message_handle, "application%2fjson");
        (void)IoTHubMessage_SetContentEncodingSystemProperty(message_handle, "utf-8");*/

      // Add custom properties to message
      // (void)IoTHubMessage_SetProperty(message_handle, "property_key", "property_value");

      LogInfo("Sending message %d to IoTHub\r\n", (int)(messages_sent + 1));
      result = IoTHubDeviceClient_LL_SendEventAsync(device_ll_handle, message_handle, send_confirm_callback, NULL);
      // The message is copied to the sdk so the we can destroy it
      IoTHubMessage_Destroy(message_handle);
      syslog(LOG_NOTICE, "message sending");

      messages_sent++;
    }

    else if (g_message_count_send_confirmations >= MESSAGE_COUNT)
    {
      // After all messages are all received stop running
      g_continueRunning = false;
    }

    IoTHubDeviceClient_LL_DoWork(device_ll_handle);
    ThreadAPI_Sleep(3);
    reset_esp_helper();

  } while (g_continueRunning);

  syslog(LOG_NOTICE, "done with sending");
}

void cleanupSDK() {
  // Clean up the iothub sdk handle
  IoTHubDeviceClient_LL_Destroy(device_ll_handle);
  // Free all the sdk subsystem
  IoTHub_Deinit();

  return;
}
/*
    メインタスク
*/
void main_task(intptr_t exinf)
{
  char	c;
  ID		tskid = TASK1;
  int_t	tskno = 1;
  ER_UINT	ercd;
  PRI		tskpri;
#ifndef TASK_LOOP
  volatile ulong_t	i;
  SYSTIM	stime1, stime2;
#endif /* TASK_LOOP */
#ifdef TOPPERS_SUPPORT_GET_UTM
  SYSUTM	utime1, utime2;
#endif /* TOPPERS_SUPPORT_GET_UTM */

  SVC_PERROR(syslog_msk_log(LOG_UPTO(LOG_INFO), LOG_UPTO(LOG_EMERG)));
  syslog(LOG_NOTICE, "Sample program starts.\n");

  /*
      シリアルポートの初期化

      システムログタスクと同じシリアルポートを使う場合など，シリアル
      ポートがオープン済みの場合にはここでE_OBJエラーになるが，支障は
      ない．
  */
//  ercd = serial_opn_por(TASK_PORTID);
//  if (ercd < 0 && MERCD(ercd) != E_OBJ) {
//    syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
//           itron_strerror(ercd), SERCD(ercd));
//  }
//  SVC_PERROR(serial_ctl_por(TASK_PORTID,
//                            (IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

#ifdef TASK_LOOP
  task_loop = TASK_LOOP;
#else /* TASK_LOOP */

#ifdef MEASURE_TWICE
  task_loop = LOOP_REF;
  SVC_PERROR(get_tim(&stime1));
  for (i = 0; i < task_loop; i++);
  SVC_PERROR(get_tim(&stime2));
#endif /* MEASURE_TWICE */

  task_loop = LOOP_REF;
  SVC_PERROR(get_tim(&stime1));
  for (i = 0; i < task_loop; i++);
  SVC_PERROR(get_tim(&stime2));
  task_loop = LOOP_REF * 400UL / (stime2 - stime1);

#endif /* TASK_LOOP */

  int j = 0;

  tft.fillScreen(TFT_BLACK); // Clear screen to navy background
  tft.setCursor(xpos, ypos);    // Set cursor near top left corner of screen

  tft.setFreeFont(FSB12);       // Select Free Serif 24 point font
  tft.println("Switch : Enter the command");
  tft.println("Pless : Command End");
  tft.println();

  while (digitalRead(WIO_5S_PRESS) != LOW) {
    if (digitalRead(WIO_5S_UP) == LOW) {
      tft.print("L");
      motor_control[j] = 'l';
      j++;
      delay(200);
    }
    else if (digitalRead(WIO_5S_DOWN) == LOW) {
      tft.print("R");
      motor_control[j] = 'r';
      j++;
      delay(200);
    }
    else if (digitalRead(WIO_5S_LEFT) == LOW) {
      tft.print("B");
      motor_control[j] = 'b';
      j++;
      delay(200);
    }
    else if (digitalRead(WIO_5S_RIGHT) == LOW) {
      tft.print("F");
      motor_control[j] = 'f';
      j++;
      delay(200);
    }
  }

  motor_control[j] = '.';
  tft.print(".");
  tft.println();

  syslog(LOG_NOTICE, "%s", motor_control);

  /*
     メインループ
  */
  int count = 0;
  while (count < j) {
    SVC_PERROR(act_tsk(TASK8));
    messages_sent = 0;
    dly_tsk(10000);
    count++;
  }

  tft.println();
  tft.println();
  tft.print("Executing End");
  tft.println();
  tft.print("Thank you!!");
  
  dly_tsk(2000);

  cleanupSDK();

  syslog(LOG_NOTICE, "Sample program ends.\n");
  SVC_PERROR(ext_ker());
  assert(0);
}
